
#include "BFSearch.h"
#include "Hashing.h"
#include "Queue.h"
#include "Maze.h"

#include <stdlib.h>

TDirection hashesToDirection(unsigned int source, unsigned int destination) {
  unsigned int sX, sY, dX, dY;
  hashReverse(source, &sX, &sY);
  hashReverse(destination, &dX, &dY);
  if (dX - sX == 1 && dY == sY) {
    return DIRECTION_RIGHT;
  }
  if (sX - dX == 1 && dY == sY) {
    return DIRECTION_LEFT;
  }
  if (dY - sY == 1 && dX == sX) {
    return DIRECTION_DOWN;
  }
  if (sY - dY == 1 && dX == sX) {
    return DIRECTION_UP;
  }
  return 0;
}

void directionToInstructions(TDirection direction, TDirection* robotDirection, TQueue* instructions) {
  // dopredu
  if (*robotDirection == direction) {
    queueAdd(instructions, INSTRUCTION_FORWARD);
    return;
  }
  // dozadu
  TDirection rotated = rotatePercept(*robotDirection, 2);
  if (rotated == direction) {
    queueAdd(instructions, INSTRUCTION_BACKWARD);
    return;
  }
  // vlavo + dopredu
  rotated = rotatePercept(*robotDirection, 3);
  if (rotated == direction) {
    *robotDirection = rotated;
    queueAdd(instructions, INSTRUCTION_TURN_LEFT);
    queueAdd(instructions, INSTRUCTION_FORWARD);
    return;
  }
  // vpravo + dopredu
  rotated = rotatePercept(*robotDirection, 1);
  if (rotated == direction) {
    *robotDirection = rotated;
    queueAdd(instructions, INSTRUCTION_TURN_RIGHT);
    queueAdd(instructions, INSTRUCTION_FORWARD);
    return;
  }
}

void wayToInstructions(TStack* way, TQueue* instructions) {
  // inicializacia
  queueInitialize(instructions);
  TDirection robotDirection = mazeGet()->robot.direction;

  // po smeroch
  while (!stackIsEmpty(way)) {
    TDirection direction = stackPop(way);
    directionToInstructions(direction, &robotDirection, instructions);
  }
}

void createWay(int finish, int *previous, TStack* way) {
  int index1 = finish, index2 = -1;
  stackInitialize(way);
  while (index1 >= 0) {
    index2 = previous[index1];
    if (index2 >= 0) {
      stackPush(way, hashesToDirection(index2, index1));
    }
    index1 = index2;
  }
}

int getNextWayXY(TStack* way, int destinationX, int destinationY) {

  // inicializacie
  THashSet visited;
  hashSetInitialize(visited);

  int previous[MAX_MAZE_SIZE * MAX_MAZE_SIZE];
  unsigned int i;
  for (i = 0; i < MAX_MAZE_SIZE * MAX_MAZE_SIZE; i++) {
    previous[i] = -1;
  }

  int finish = -1;
  TQueue search;
  queueInitialize(&search);
  queueAdd(&search, hashXY(mazeGet()->robot.x, mazeGet()->robot.y));

  // prehladavanie
  while (!queueIsEmpty(&search)) {
    unsigned int x, y, current = queuePop(&search);
    hashSetAdd(visited, current);
    hashReverse(current, &x, &y);
    TCellInfo walls = mazeCell(x, y);

    // zistim ci som neni v "cieli"
    if ((destinationX >= 0 && destinationY >= 0 && x == destinationX && y == destinationY) ||               // ciel je definovany ako X, Y
        (destinationX < 0 && destinationY < 0 && ((walls & KNOWN_ALL_DIRECTIONS) != KNOWN_ALL_DIRECTIONS))) // ciel je polciko, ktore nema znema vsetky steny
    {
      finish = current;
      break;
    } else {
      // niesom - pridanie susedov
      // horny sused
      unsigned int hash;
      if (((walls & DIRECTION_UP) == 0) && ((walls & KNOWN_DIRECTION_UP) > 0)) {
        hash = hashXY(x, y - 1);
        if (!hashSetContains(visited, hash)) {
          hashSetAdd(visited, hash);
          queueAdd(&search, hash);
          previous[hash] = current;
        }
      }
      // dolny sused
      if (((walls & DIRECTION_DOWN) == 0) && ((walls & KNOWN_DIRECTION_DOWN) > 0)) {
        hash = hashXY(x, y + 1);
        if (!hashSetContains(visited, hash)) {
          hashSetAdd(visited, hash);
          queueAdd(&search, hash);
          previous[hash] = current;
        }
      }
      // pravy sused
      if (((walls & DIRECTION_RIGHT) == 0) && ((walls & KNOWN_DIRECTION_RIGHT) > 0)) {
        hash = hashXY(x + 1, y);
        if (!hashSetContains(visited, hash)) {
          hashSetAdd(visited, hash);
          queueAdd(&search, hash);
          previous[hash] = current;
        }
      }
      // lavy sused
      if (((walls & DIRECTION_LEFT) == 0) && ((walls & KNOWN_DIRECTION_LEFT) > 0)) {
        hash = hashXY(x - 1, y);
        if (!hashSetContains(visited, hash)) {
          hashSetAdd(visited, hash);
          queueAdd(&search, hash);
          previous[hash] = current;
        }
      }
    }
  }
  queueFreeItems(&search);

  // vysledok - nasiel sa
  if (finish >= 0) {
    // vytvorenie cesty zo spajaneho zoznamu, ktoreho koniec je pointer finish
    createWay(finish, previous, way);
  }
  return finish >= 0;
}
