
#include "Maze.h"

TMaze maze;

void robotInitialize() {
  maze.robot.x = 0;
  maze.robot.y = 0;
  maze.robot.direction = DIRECTION_DOWN;
}

void mazeInitialize() {
  robotInitialize();
  unsigned int i, j;
  for (i = 0; i < MAX_MAZE_SIZE - 1; i++) {
    for (j = 0; j < MAX_MAZE_SIZE; j++) {
      maze.walls[i][j] = 0;
    }
  }
}

TCellInfo mazeCellLeft(unsigned int x, unsigned int y) {
  TCellInfo result = 0;
  if (x == 0) {
    // krajny pripad
    result |= KNOWN_DIRECTION_LEFT | DIRECTION_LEFT;
  } else {
    // pozrem sa co mam ulozene v poli
    unsigned char leftWall = maze.walls[x - 1][y] & 0x0F;
    if (leftWall != WALL_NOT_KNOWN) {
      result |= KNOWN_DIRECTION_LEFT;
    }
    if (leftWall == WALL_IS_THERE) {
      result |= DIRECTION_LEFT;
    }
  }
  return result;
}

TCellInfo mazeCellRight(unsigned int x, unsigned int y) {
  TCellInfo result = 0;
  if (x == MAX_MAZE_SIZE - 1) {
    // krajny pripad
    result |= KNOWN_DIRECTION_RIGHT | DIRECTION_RIGHT;
  } else {
    // pozrem sa co mam ulozene v poli
    unsigned char rightWall = maze.walls[x][y] & 0x0F;
    if (rightWall != WALL_NOT_KNOWN) {
      result |= KNOWN_DIRECTION_RIGHT;
    }
    if (rightWall == WALL_IS_THERE) {
      result |= DIRECTION_RIGHT;
    }
  }
  return result;
}

TCellInfo mazeCellUp(unsigned int x, unsigned int y) {
  TCellInfo result = 0;
  if (y == 0) {
    // krajny pripad
    result |= KNOWN_DIRECTION_UP | DIRECTION_UP;
  } else {
    // pozrem sa co mam ulozene v poli
    unsigned char topWall = maze.walls[y - 1][x] >> 4;
    if (topWall != WALL_NOT_KNOWN) {
      result |= KNOWN_DIRECTION_UP;
    }
    if (topWall == WALL_IS_THERE) {
      result |= DIRECTION_UP;
    }
  }
  return result;
}

TCellInfo mazeCellDown(unsigned int x, unsigned int y) {
  TCellInfo result = 0;
  if (y == MAX_MAZE_SIZE - 1) {
    // krajny pripad
    result |= KNOWN_DIRECTION_DOWN | DIRECTION_DOWN;
  } else {
    // pozrem sa co mam ulozene v poli
    unsigned char bottomWall = maze.walls[y][x] >> 4;
    if (bottomWall != WALL_NOT_KNOWN) {
      result |= KNOWN_DIRECTION_DOWN;
    }
    if (bottomWall == WALL_IS_THERE) {
      result |= DIRECTION_DOWN;
    }
  }
  return result;
}

TCellInfo mazeCell(unsigned int x, unsigned int y) {
  // mimo plochy vieme o tom ze vsade su same steny
  if (x < 0 || x >= MAX_MAZE_SIZE || y < 0 || y >= MAX_MAZE_SIZE) {
    return 0xFF;
  }

  // otestujem vsetky umiestnenia stien
  return
    mazeCellLeft(x, y)  |
    mazeCellDown(x, y)   |
    mazeCellRight(x, y) |
    mazeCellUp(x, y);
}

TCellInfo rotatePercept(TCellInfo percept, unsigned int rotation) {
  TCellInfo temp = (percept << rotation) | (percept >> (8 - rotation));
  return ((temp >> 4) | temp) & 0x0F;
}

TCellInfo orientPercept(TCellInfo percept) {
  if (maze.robot.direction & DIRECTION_DOWN) {
    return rotatePercept(percept, 2);
  }
  if (maze.robot.direction & DIRECTION_LEFT) {
    return rotatePercept(percept, 3);
  }
  if (maze.robot.direction & DIRECTION_RIGHT) {
    return rotatePercept(percept, 1);
  }
  return percept;
}

TWallInfo isWall(TCellInfo percept, TDirection direction) {
  return (percept & direction) ? WALL_IS_THERE : WALL_IS_NOT_THERE;
}

void addPercept(TCellInfo percept) {
  TCellInfo oPercept = orientPercept(percept);
  unsigned int x = maze.robot.x;
  unsigned int y = maze.robot.y;

  // lava stena
  if (x > 0) {
    if ((maze.walls[x - 1][y] & 0x0F) == WALL_NOT_KNOWN) {
      maze.walls[x - 1][y] = isWall(oPercept, DIRECTION_LEFT) | (maze.walls[x - 1][y] & 0xF0);
    }
  }
  // prava stena
  if (x < MAX_MAZE_SIZE - 1) {
    if ((maze.walls[x][y] & 0x0F) == WALL_NOT_KNOWN) {
      maze.walls[x][y] = isWall(oPercept, DIRECTION_RIGHT) | (maze.walls[x][y] & 0xF0);
    }
  }
  // horna stena
  if (y > 0) {
    if ((maze.walls[y - 1][x] >> 4) == WALL_NOT_KNOWN) {
      maze.walls[y - 1][x] = (isWall(oPercept, DIRECTION_UP) << 4) | (maze.walls[y - 1][x] & 0x0F);
    }
  }
  // dolna stena
  if (y < MAX_MAZE_SIZE - 1) {
    if ((maze.walls[y][x] >> 4) == WALL_NOT_KNOWN) {
      maze.walls[y][x] = (isWall(oPercept, DIRECTION_DOWN) << 4) | (maze.walls[y][x] & 0x0F);
    }
  }
}

TMaze* mazeGet() {
  return &maze;
}

int getFinishCoords(unsigned int *x, unsigned int *y) {
  unsigned int i, j;
  for (i = 0; i < MAX_MAZE_SIZE - 1; i++) {
    for (j = 0; j < MAX_MAZE_SIZE - 1; j++) {
      TCellInfo info1 = mazeCell(i, j);
      TCellInfo info2 = mazeCell(i + 1, j + 1);
      int bool1 =
        ((info1 & KNOWN_DIRECTION_RIGHT) > 0) && ((info1 & KNOWN_DIRECTION_DOWN) > 0) &&
        ((info1 & DIRECTION_RIGHT) == 0) && ((info1 & DIRECTION_DOWN) == 0);
      int bool2 =
        ((info2 & KNOWN_DIRECTION_LEFT) > 0) && ((info2 & KNOWN_DIRECTION_UP) > 0) &&
        ((info2 & DIRECTION_LEFT) == 0) && ((info2 & DIRECTION_UP) == 0);
      if (bool1 && bool2) {
        *x = i;
        *y = j;
        return 1;
      }
    }
  }
  return 0;
}
