#ifndef _CARLDRIVE__H
#define _CARLDRIVE__H

#define PI 3.1415926
#define DIAMETER 41.35
#define CIRCUMFERENCE (DIAMETER * PI)
#define MOTOR_STEPS_PER_REVOLUTION 20
#define MOTOR_TO_WHEEL_GEAR_RATIO 50
#define STEPS_PER_REVOLUTION (MOTOR_STEPS_PER_REVOLUTION * MOTOR_TO_WHEEL_GEAR_RATIO)
#define STEP_PER_D (float)(STEPS_PER_REVOLUTION / CIRCUMFERENCE)
#define MM_TO_STEPS(mm) (int)(mm * STEP_PER_D)

#define WHEEL_SEPARATION 54.33
#define STEPS_FOR_FULL_CIRCLE (int)((WHEEL_SEPARATION / DIAMETER) * STEPS_PER_REVOLUTION)

/**
 *      Moves the robot at a certain speed.
 * @param lspeed the speed at which to move the left motor (0 - 1000; positive 
 *        for forward, negative for backward). This speed is in steps / 
 *        second.
 * @param rspeed the speed at which to move the right motor, same legal values 
 *        as lspeed.
 * @param stepcount the number of steps at which to move. Movement will cease 
 *        once the faster motor has moved this many steps. Use MM_TO_STEPS(mm) 
 *        to move a distance, or some multiple (or fraction) of 
 *        STEPS_FOR_FULL_CIRCLE to move in a circle.
 */
void drive(int lspeed, int rspeed, int stepcount);
#endif
