
#include <p30f6014A.h>
#include <stdio.h>

#include "../library/motor_led/e_epuck_ports.h"
#include "../library/motor_led/e_init_port.h"
#include "../library/motor_led/e_led.h"
#include "../library/motor_led/e_motors.h"

#include "../library/a_d/e_prox.h"
#include "Actuators.h"
#include "Maze.h"
#include "Stack.h"
#include "Queue.h"
#include "BFSearch.h"

int getselector() {
  return SELECTOR0 + 2*SELECTOR1 + 4*SELECTOR2 + 8*SELECTOR3;
}

void disco() {
  unsigned int ledIndex = 0;
  while (1) {
  	long i;
  	for(i = 0; i < 250000; i++) {
  		asm("NOP");
  	}
  	e_set_led(ledIndex, 2); 
    ledIndex++;
    ledIndex %= 8;
  }
}

void sensors() {
  int value;
  	while(1)
  	{
  		long i;
        int j;
        for (j=0; j<8; j++) {
	  		value = e_get_prox(j);
	  		if(value > 135)	
	  			e_set_led((j+1) %8 , 1); 
	  		else
	  			e_set_led((j+1) %8, 0); 
        } 
  		for(i=0; i<100000; i++) { asm("nop"); }
  	}
}

void executeInstructions(TQueue *queue) {
  while (!queueIsEmpty(queue)) {
    switch (queuePop(queue)) {
      case INSTRUCTION_FORWARD:
        move(18);
        switch (mazeGet()->robot.direction) {
          case DIRECTION_UP:
            mazeGet()->robot.y--;
          break;
		  case DIRECTION_DOWN:
            mazeGet()->robot.y++;
          break;
		  case DIRECTION_LEFT:
            mazeGet()->robot.x--;
          break;
		  case DIRECTION_RIGHT:
            mazeGet()->robot.x++;
          break;
        }        
      break;
      case INSTRUCTION_BACKWARD:
        move(-18);
        switch (mazeGet()->robot.direction) {
          case DIRECTION_UP:
            mazeGet()->robot.y++;
          break;
		  case DIRECTION_DOWN:
            mazeGet()->robot.y--;
          break;
		  case DIRECTION_LEFT:
            mazeGet()->robot.x++;
          break;
		  case DIRECTION_RIGHT:
            mazeGet()->robot.x--;
          break;
        }        
      break; 
      case INSTRUCTION_TURN_LEFT:
        turn(270);
        mazeGet()->robot.direction = rotatePercept(mazeGet()->robot.direction, 3);
      break;
      case INSTRUCTION_TURN_RIGHT:
        turn(90);
        mazeGet()->robot.direction = rotatePercept(mazeGet()->robot.direction, 1);
      break; 
    }
  }
}

void bludisko() {
  mazeInitialize();
  TStack way;
  TQueue instructions;

  // prehladanie
  long i;
  for (i = 0; i < 500000; i++) { asm("nop"); }

  addPercept(getPercept());
  while (getNextWay(&way)) {
    wayToInstructions(&way, &instructions);
    executeInstructions(&instructions);
    addPercept(getPercept());
  }
  
  // na start
  if (getNextWayXY(&way, 0, 0)) {
    wayToInstructions(&way, &instructions);
    executeInstructions(&instructions);
  }

  for (i = 0; i < 5000000; i++) { asm("nop"); }

  // do ciela
  unsigned int x, y;
  if (getFinishCoords(&x, &y)) {
    if (getNextWayXY(&way, x, y)) {
      wayToInstructions(&way, &instructions);
      executeInstructions(&instructions);
    }
  }
}

int main() {

  // inicializacia
  e_init_port();
  e_init_motors();
  e_init_prox();
  	
  // v pripade zapnutia resetnem (kvoli problemom s niektorymi robotmi)
  if (RCONbits.POR) {
	RCONbits.POR = 0;
	__asm__ volatile ("reset");
  }

  e_set_body_led(1); 
 
  int selector=getselector();	
  if (selector==0) {
	bludisko();
  } else if (selector==1) {
	sensors();
  } else if (selector==2) {
	getPercept();
  } else if (selector==3) {
    move(16);
  } else if (selector==4) {
    turn(180);
  } else {
	disco();
  }   
  while(1);
  return 0;	
}

