#include "CuTest.h"
#include "../BFSearch.h"
#include "../Stack.h"
#include "../Maze.h"

TDirection hashesToDirection(unsigned int source, unsigned int destination);
void directionToInstructions(TDirection direction, TDirection* robotDirection, TQueue* instructions);

void createTestMaze();

void BFSearchTestHashesToDirection(CuTest *tc) {
  CuAssertIntEquals(tc, DIRECTION_RIGHT, hashesToDirection(0, 1));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, hashesToDirection(4, 5));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, hashesToDirection(52, 53));
  CuAssertIntEquals(tc, DIRECTION_LEFT, hashesToDirection(1, 0));
  CuAssertIntEquals(tc, DIRECTION_LEFT, hashesToDirection(7, 6));
  CuAssertIntEquals(tc, DIRECTION_LEFT, hashesToDirection(50, 49));
  CuAssertIntEquals(tc, DIRECTION_DOWN, hashesToDirection(0, MAX_MAZE_SIZE));
  CuAssertIntEquals(tc, DIRECTION_DOWN, hashesToDirection(7, MAX_MAZE_SIZE + 7));
  CuAssertIntEquals(tc, DIRECTION_DOWN, hashesToDirection(50, MAX_MAZE_SIZE + 50));
  CuAssertIntEquals(tc, DIRECTION_UP, hashesToDirection(MAX_MAZE_SIZE + 1, 1));
  CuAssertIntEquals(tc, DIRECTION_UP, hashesToDirection(MAX_MAZE_SIZE + 7, 7));
  CuAssertIntEquals(tc, DIRECTION_UP, hashesToDirection(MAX_MAZE_SIZE + 42, 42));
  CuAssertIntEquals(tc, 0, hashesToDirection(MAX_MAZE_SIZE, MAX_MAZE_SIZE - 1));
}

void BFSearchTestPath(CuTest *tc) {
  mazeInitialize();
  addPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_LEFT);
  TStack way;

  // prva cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_DOWN, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 0;
  mazeGet()->robot.y = 1;
  addPercept(DIRECTION_RIGHT | DIRECTION_UP);

  // druha cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 1;
  mazeGet()->robot.y = 1;
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  addPercept(DIRECTION_LEFT);

  // tretia cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_DOWN, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 1;
  mazeGet()->robot.y = 2;
  mazeGet()->robot.direction = DIRECTION_DOWN;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT | DIRECTION_UP);

  // stvrta cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 2;
  mazeGet()->robot.y = 1;
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);

  // piata cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 3;
  mazeGet()->robot.y = 1;
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  addPercept(DIRECTION_UP | DIRECTION_RIGHT);

  // siesta cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 3;
  mazeGet()->robot.y = 0;
  mazeGet()->robot.direction = DIRECTION_UP;
  addPercept(DIRECTION_UP | DIRECTION_RIGHT);

  // siedma cesta
  CuAssertTrue(tc, getNextWay(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_LEFT, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 2;
  mazeGet()->robot.y = 0;
  mazeGet()->robot.direction = DIRECTION_LEFT;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);

  CuAssertTrue(tc, !getNextWay(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));
}

void BFSearchTestPathXY(CuTest *tc) {
  mazeInitialize();
  createTestMaze();
  mazeGet()->robot.x = 0;
  mazeGet()->robot.y = 0;
  TStack way;

  // cesta 1
  CuAssertTrue(tc, getNextWayXY(&way, 2, 2));
  unsigned int i;
  for (i = 0; i < 4; i++) {
    CuAssertTrue(tc, !stackIsEmpty(&way));
    CuAssertIntEquals(tc, DIRECTION_DOWN, stackPop(&way));
  }
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));

  // cesta 2
  mazeGet()->robot.x = 2;
  mazeGet()->robot.y = 2;
  CuAssertTrue(tc, getNextWayXY(&way, 3, 4));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_LEFT, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_DOWN, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_DOWN, stackPop(&way));
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_LEFT, stackPop(&way));
  for (i = 0; i < 3; i++) {
    CuAssertTrue(tc, !stackIsEmpty(&way));
    CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  }
  for (i = 0; i < 3; i++) {
    CuAssertTrue(tc, !stackIsEmpty(&way));
    CuAssertIntEquals(tc, DIRECTION_RIGHT, stackPop(&way));
  }
  for (i = 0; i < 3; i++) {
    CuAssertTrue(tc, !stackIsEmpty(&way));
    CuAssertIntEquals(tc, DIRECTION_DOWN, stackPop(&way));
  }
  CuAssertTrue(tc, stackIsEmpty(&way));
  mazeGet()->robot.x = 3;
  mazeGet()->robot.y = 4;

  // cesta 3
  CuAssertTrue(tc, getNextWayXY(&way, 0, 0));
  for (i = 0; i < 3; i++) {
    CuAssertTrue(tc, !stackIsEmpty(&way));
    CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  }
  for (i = 0; i < 3; i++) {
    CuAssertTrue(tc, !stackIsEmpty(&way));
    CuAssertIntEquals(tc, DIRECTION_LEFT, stackPop(&way));
  }
  CuAssertTrue(tc, !stackIsEmpty(&way));
  CuAssertIntEquals(tc, DIRECTION_UP, stackPop(&way));
  CuAssertTrue(tc, stackIsEmpty(&way));

  // neexistujuca cesta
  CuAssertTrue(tc, !getNextWayXY(&way, 6, 5));
}

void BFSearchTestDirToInstruct(CuTest* tc) {
  // inicializacia
  TDirection robot = DIRECTION_UP;
  TQueue queue;
  queueInitialize(&queue);

  // rovno
  directionToInstructions(DIRECTION_UP, &robot, &queue);
  CuAssertIntEquals(tc, DIRECTION_UP, robot);
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&queue));
  CuAssertTrue(tc, queueIsEmpty(&queue));

  // dozadu
  directionToInstructions(DIRECTION_DOWN, &robot, &queue);
  CuAssertIntEquals(tc, DIRECTION_UP, robot);
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_BACKWARD, queuePop(&queue));
  CuAssertTrue(tc, queueIsEmpty(&queue));

  // vlavo
  directionToInstructions(DIRECTION_LEFT, &robot, &queue);
  CuAssertIntEquals(tc, DIRECTION_LEFT, robot);
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_LEFT, queuePop(&queue));
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&queue));
  CuAssertTrue(tc, queueIsEmpty(&queue));

  // rovno
  directionToInstructions(DIRECTION_LEFT, &robot, &queue);
  CuAssertIntEquals(tc, DIRECTION_LEFT, robot);
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&queue));
  CuAssertTrue(tc, queueIsEmpty(&queue));

  // vpravo
  directionToInstructions(DIRECTION_UP, &robot, &queue);
  CuAssertIntEquals(tc, DIRECTION_UP, robot);
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_RIGHT, queuePop(&queue));
  CuAssertTrue(tc, !queueIsEmpty(&queue));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&queue));
  CuAssertTrue(tc, queueIsEmpty(&queue));
}

void BFSearchTestWayToInstruct(CuTest* tc) {
  // initialize
  mazeInitialize();
  TStack way;
  stackInitialize(&way);
  stackPush(&way, DIRECTION_LEFT);
  stackPush(&way, DIRECTION_UP);
  stackPush(&way, DIRECTION_RIGHT);
  stackPush(&way, DIRECTION_RIGHT);
  stackPush(&way, DIRECTION_UP);
  stackPush(&way, DIRECTION_DOWN);
  stackPush(&way, DIRECTION_RIGHT);
  stackPush(&way, DIRECTION_DOWN);

  // prevod
  TQueue instructions;
  queueInitialize(&instructions);
  wayToInstructions(&way, &instructions);

  // test vysledku
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));    CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_LEFT, queuePop(&instructions));  CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));    CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_RIGHT, queuePop(&instructions)); CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));    CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_BACKWARD, queuePop(&instructions));   CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_LEFT, queuePop(&instructions));  CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));    CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));    CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_LEFT, queuePop(&instructions));  CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));    CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_TURN_LEFT, queuePop(&instructions));  CuAssertTrue(tc, !queueIsEmpty(&instructions));
  CuAssertIntEquals(tc, INSTRUCTION_FORWARD, queuePop(&instructions));
  CuAssertTrue(tc, queueIsEmpty(&instructions));
}

CuSuite* BFSearchTestsGetSuite() {
  CuSuite* suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, BFSearchTestHashesToDirection);
  SUITE_ADD_TEST(suite, BFSearchTestPath);
  SUITE_ADD_TEST(suite, BFSearchTestPathXY);
  SUITE_ADD_TEST(suite, BFSearchTestDirToInstruct);
  SUITE_ADD_TEST(suite, BFSearchTestWayToInstruct);
  return suite;
}
