#include "CuTest.h"
#include "../Maze.h"
#include "../Hashing.h"

#define TEST_HASHING_SET_TOTAL 10

void HashingTestHashFunc(CuTest *tc) {
  unsigned int i1, i2, j1, j2;
  // test kazdej dvojice s kazdou, ci sa funkcia sprava tak, ako by mala - nedava rovnake hodnoty pre rozne cisla
  for (i1 = 0; i1 < MAX_MAZE_SIZE; i1++) {
    for (j1 = 0; j1 < MAX_MAZE_SIZE; j1++) {
      for (i2 = 0; i2 < MAX_MAZE_SIZE; i2++) {
        for (j2 = 0; j2 < MAX_MAZE_SIZE; j2++) {
          unsigned int H1 = hashXY(i1, j1);
          unsigned int H2 = hashXY(i2, j2);
          if (i1 == i2 && j1 == j2) {
            CuAssertIntEquals(tc, H1, H2);
          } else {
            CuAssertTrue(tc, H1 != H2);
          }
        }
      }
    }
  }
}

void HashingTestHashReverse(CuTest *tc) {
  unsigned int i, j;
  // test pomocou funkcie na hashovanie
  for (i = 0; i < MAX_MAZE_SIZE; i++) {
    for (j = 0; j < MAX_MAZE_SIZE; j++) {
      unsigned int hash = hashXY(i, j);
      unsigned int x, y;
      hashReverse(hash, &x, &y);
      CuAssertIntEquals(tc, i, x);
      CuAssertIntEquals(tc, j, y);
    }
  }
}

void HashingTestSetInit(CuTest *tc) {
  THashSet set;
  hashSetInitialize(set);
  unsigned int i, j;
  for (i = 0; i < MAX_MAZE_SIZE; i++) {
    for (j = 0; j < MAX_MAZE_SIZE; j++) {
      CuAssertTrue(tc, !hashSetContains(set, hashXY(i, j)));
    }
  }
}

int inArray(int *array, int value) {
  int appearance = 0;
  unsigned int i;
  for (i = 0; i < TEST_HASHING_SET_TOTAL; i++) {
    if (array[i] == value) {
      appearance++;
    }
  }
  return appearance > 0;
}

void HashingTestSet(CuTest *tc) {
  int hashes[] = {5, 6, 9, 12, 52, 102, 3, 4, 9, 0};

  THashSet set;
  hashSetInitialize(set);

  // nahadzem hashe z pola
  unsigned int i, j;
  for (i = 0; i < TEST_HASHING_SET_TOTAL; i++) {
    hashSetAdd(set, hashes[i]);
  }

  // otestujem ci tam su
  for (i = 0; i < MAX_MAZE_SIZE; i++) {
    for (j = 0; j < MAX_MAZE_SIZE; j++) {
      unsigned int hash = hashXY(i, j);
      int contains = hashSetContains(set, hash);
      CuAssertTrue(tc, contains == inArray(hashes, hash));
    }
  }
}

CuSuite* HashingTestsGetSuite() {
  CuSuite* suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, HashingTestHashFunc);
  SUITE_ADD_TEST(suite, HashingTestHashReverse);
  SUITE_ADD_TEST(suite, HashingTestSetInit);
  SUITE_ADD_TEST(suite, HashingTestSet);
  return suite;
}
