#include "CuTest.h"
#include "../Maze.h"

// hlavicky "private" funkcii, pre testovanie
TCellInfo orientPercept(TCellInfo percept);

void MazeTestInit(CuTest *tc) {
  mazeInitialize();
  unsigned int i, j;
  for (i = 0; i < MAX_MAZE_SIZE - 1; i++) {
    for (j = 0; j < MAX_MAZE_SIZE; j++) {
      CuAssertIntEquals(tc, 0, mazeGet()->walls[i][j]);
    }
  }
}

void MazeTestInitWallCorners(CuTest *tc) {
  mazeInitialize();
  TCellInfo leftTop     = KNOWN_DIRECTION_LEFT  | KNOWN_DIRECTION_UP   | DIRECTION_LEFT  | DIRECTION_UP;
  TCellInfo rightTop    = KNOWN_DIRECTION_RIGHT | KNOWN_DIRECTION_UP   | DIRECTION_RIGHT | DIRECTION_UP;
  TCellInfo leftBottom  = KNOWN_DIRECTION_LEFT  | KNOWN_DIRECTION_DOWN | DIRECTION_LEFT  | DIRECTION_DOWN;
  TCellInfo rightBottom = KNOWN_DIRECTION_RIGHT | KNOWN_DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_DOWN;
  CuAssertIntEquals(tc, leftTop, mazeCell(0, 0));
  CuAssertIntEquals(tc, rightTop, mazeCell(MAX_MAZE_SIZE - 1, 0));
  CuAssertIntEquals(tc, leftBottom, mazeCell(0, MAX_MAZE_SIZE - 1));
  CuAssertIntEquals(tc, rightBottom, mazeCell(MAX_MAZE_SIZE - 1, MAX_MAZE_SIZE - 1));
}

void MazeTestInitWallSides(CuTest *tc) {
  mazeInitialize();
  TCellInfo left   = KNOWN_DIRECTION_LEFT  | DIRECTION_LEFT;
  TCellInfo right  = KNOWN_DIRECTION_RIGHT | DIRECTION_RIGHT;
  TCellInfo top    = KNOWN_DIRECTION_UP    | DIRECTION_UP;
  TCellInfo bottom = KNOWN_DIRECTION_DOWN  | DIRECTION_DOWN;
  unsigned int i;
  for (i = 1; i < MAX_MAZE_SIZE - 1; i++) {
    CuAssertIntEquals(tc, left, mazeCell(0, i));
    CuAssertIntEquals(tc, top,  mazeCell(i, 0));
    CuAssertIntEquals(tc, right, mazeCell(MAX_MAZE_SIZE - 1, i));
    CuAssertIntEquals(tc, bottom,  mazeCell(i, MAX_MAZE_SIZE - 1));
  }
}

void MazeTestInitWallCenter(CuTest *tc) {
  mazeInitialize();
  unsigned int i, j;
  for (i = 1; i < MAX_MAZE_SIZE - 1; i++) {
    for (j = 1; j < MAX_MAZE_SIZE - 1; j++) {
      CuAssertIntEquals(tc, 0, mazeCell(i, j));
    }
  }
}

void MazeTestPerceptRotation180(CuTest *tc) {
  CuAssertIntEquals(tc, DIRECTION_UP, rotatePercept(DIRECTION_DOWN, 2));
  CuAssertIntEquals(tc, DIRECTION_DOWN, rotatePercept(DIRECTION_UP, 2));
  CuAssertIntEquals(tc, DIRECTION_UP | DIRECTION_DOWN, rotatePercept(DIRECTION_UP | DIRECTION_DOWN, 2));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_DOWN, rotatePercept(DIRECTION_LEFT | DIRECTION_UP, 2));
  CuAssertIntEquals(tc, DIRECTION_LEFT | DIRECTION_DOWN, rotatePercept(DIRECTION_RIGHT | DIRECTION_UP, 2));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_UP | DIRECTION_LEFT | DIRECTION_DOWN, rotatePercept(DIRECTION_RIGHT | DIRECTION_UP | DIRECTION_LEFT | DIRECTION_DOWN, 2));
  CuAssertIntEquals(tc, 0, rotatePercept(0, 2));
}

void MazeTestPerceptRotation90(CuTest *tc) {
  CuAssertIntEquals(tc, DIRECTION_LEFT, rotatePercept(DIRECTION_DOWN, 1));
  CuAssertIntEquals(tc, DIRECTION_RIGHT, rotatePercept(DIRECTION_UP, 1));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_LEFT, rotatePercept(DIRECTION_UP | DIRECTION_DOWN, 1));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_UP, rotatePercept(DIRECTION_LEFT | DIRECTION_UP, 1));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_DOWN, rotatePercept(DIRECTION_RIGHT | DIRECTION_UP, 1));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_UP | DIRECTION_LEFT | DIRECTION_DOWN, rotatePercept(DIRECTION_RIGHT | DIRECTION_UP | DIRECTION_LEFT | DIRECTION_DOWN, 1));
  CuAssertIntEquals(tc, 0, rotatePercept(0, 1));
}

void MazeTestPerceptRotation_90(CuTest *tc) {
  CuAssertIntEquals(tc, DIRECTION_RIGHT, rotatePercept(DIRECTION_DOWN, 3));
  CuAssertIntEquals(tc, DIRECTION_LEFT, rotatePercept(DIRECTION_UP, 3));
  CuAssertIntEquals(tc, DIRECTION_LEFT | DIRECTION_RIGHT, rotatePercept(DIRECTION_UP | DIRECTION_DOWN, 3));
  CuAssertIntEquals(tc, DIRECTION_LEFT | DIRECTION_DOWN, rotatePercept(DIRECTION_LEFT | DIRECTION_UP, 3));
  CuAssertIntEquals(tc, DIRECTION_LEFT | DIRECTION_UP, rotatePercept(DIRECTION_RIGHT | DIRECTION_UP, 3));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_UP | DIRECTION_LEFT | DIRECTION_DOWN, rotatePercept(DIRECTION_RIGHT | DIRECTION_UP | DIRECTION_LEFT | DIRECTION_DOWN, 3));
  CuAssertIntEquals(tc, 0, rotatePercept(0, -1));
}

void MazeTestPerceptOrientationDown(CuTest *tc) {
  mazeInitialize();
  CuAssertIntEquals(tc, DIRECTION_UP, orientPercept(DIRECTION_DOWN));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_UP, orientPercept(DIRECTION_DOWN | DIRECTION_UP));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_DOWN, orientPercept(DIRECTION_UP | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_RIGHT | DIRECTION_UP, orientPercept(DIRECTION_DOWN | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_LEFT | DIRECTION_UP, orientPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_UP));
  CuAssertIntEquals(tc, 0, orientPercept(0));
}

void MazeTestPerceptOrientationUp(CuTest *tc) {
  mazeInitialize();
  mazeGet()->robot.direction = DIRECTION_UP;
  CuAssertIntEquals(tc, DIRECTION_DOWN, orientPercept(DIRECTION_DOWN));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_UP, orientPercept(DIRECTION_DOWN | DIRECTION_UP));
  CuAssertIntEquals(tc, DIRECTION_UP | DIRECTION_LEFT, orientPercept(DIRECTION_UP | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_LEFT, orientPercept(DIRECTION_DOWN | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_UP, orientPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_UP));
  CuAssertIntEquals(tc, 0, orientPercept(0));
}

void MazeTestPerceptOrientationLeft(CuTest *tc) {
  mazeInitialize();
  mazeGet()->robot.direction = DIRECTION_LEFT;
  CuAssertIntEquals(tc, DIRECTION_RIGHT, orientPercept(DIRECTION_DOWN));
  CuAssertIntEquals(tc, DIRECTION_LEFT | DIRECTION_RIGHT, orientPercept(DIRECTION_DOWN | DIRECTION_UP));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_LEFT, orientPercept(DIRECTION_UP | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_RIGHT, orientPercept(DIRECTION_DOWN | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_UP | DIRECTION_RIGHT | DIRECTION_LEFT, orientPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_UP));
  CuAssertIntEquals(tc, 0, orientPercept(0));
}

void MazeTestPerceptOrientationRight(CuTest *tc) {
  mazeInitialize();
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  CuAssertIntEquals(tc, DIRECTION_LEFT, orientPercept(DIRECTION_DOWN));
  CuAssertIntEquals(tc, DIRECTION_LEFT | DIRECTION_RIGHT, orientPercept(DIRECTION_DOWN | DIRECTION_UP));
  CuAssertIntEquals(tc, DIRECTION_UP | DIRECTION_RIGHT, orientPercept(DIRECTION_UP | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_UP | DIRECTION_LEFT, orientPercept(DIRECTION_DOWN | DIRECTION_LEFT));
  CuAssertIntEquals(tc, DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_LEFT, orientPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_UP));
  CuAssertIntEquals(tc, 0, orientPercept(0));
}

void MazeTestAddingPercept(CuTest *tc) {
  mazeInitialize();
  addPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_LEFT);
  mazeGet()->robot.y++;
  addPercept(DIRECTION_RIGHT | DIRECTION_UP);
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  mazeGet()->robot.x++;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);

  TCellInfo allKnown = KNOWN_ALL_DIRECTIONS;
  TCellInfo uldKnown = KNOWN_DIRECTION_LEFT | KNOWN_DIRECTION_UP | KNOWN_DIRECTION_DOWN;

  CuAssertIntEquals(tc, allKnown | DIRECTION_LEFT | DIRECTION_RIGHT | DIRECTION_UP, mazeCell(0, 0));
  CuAssertIntEquals(tc, allKnown | DIRECTION_LEFT | DIRECTION_DOWN, mazeCell(0, 1));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_LEFT | KNOWN_DIRECTION_UP | DIRECTION_LEFT | DIRECTION_UP, mazeCell(0, 2));
  CuAssertIntEquals(tc, uldKnown | DIRECTION_LEFT | DIRECTION_UP | DIRECTION_DOWN, mazeCell(1, 0));
  CuAssertIntEquals(tc, allKnown | DIRECTION_UP | DIRECTION_DOWN, mazeCell(1, 1));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_UP | DIRECTION_UP, mazeCell(1, 2));
}

void MazeTestAddingPercept2(CuTest *tc) {
  mazeInitialize();
  addPercept(DIRECTION_DOWN | DIRECTION_RIGHT | DIRECTION_UP);
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  mazeGet()->robot.x++;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_LEFT | DIRECTION_UP);
  mazeGet()->robot.direction = DIRECTION_DOWN;
  mazeGet()->robot.y++;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);
  mazeGet()->robot.y++;
  addPercept(DIRECTION_UP);
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  mazeGet()->robot.x++;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);
  mazeGet()->robot.direction = DIRECTION_LEFT;
  mazeGet()->robot.x -= 2;
  addPercept(DIRECTION_LEFT | DIRECTION_RIGHT);
  mazeGet()->robot.x--;
  addPercept(DIRECTION_LEFT | DIRECTION_UP);
  mazeGet()->robot.direction = DIRECTION_UP;
  mazeGet()->robot.y--;
  addPercept(DIRECTION_LEFT | DIRECTION_UP);
  mazeGet()->robot.direction = DIRECTION_RIGHT;
  mazeGet()->robot.x++;
  addPercept(DIRECTION_LEFT | DIRECTION_UP | DIRECTION_RIGHT);

  TCellInfo allKnown = KNOWN_ALL_DIRECTIONS;
  TCellInfo LDU = DIRECTION_LEFT  | DIRECTION_DOWN | DIRECTION_UP;
  TCellInfo DUR = DIRECTION_RIGHT | DIRECTION_DOWN | DIRECTION_UP;
  TCellInfo LU  = DIRECTION_LEFT  | DIRECTION_UP;
  TCellInfo DU  = DIRECTION_DOWN  | DIRECTION_UP;
  TCellInfo LD  = DIRECTION_LEFT  | DIRECTION_DOWN;
  TCellInfo LR  = DIRECTION_LEFT  | DIRECTION_RIGHT;
  TCellInfo UR  = DIRECTION_UP    | DIRECTION_RIGHT;

  CuAssertIntEquals(tc, allKnown | LDU, mazeCell(0, 0));
  CuAssertIntEquals(tc, allKnown | LU,  mazeCell(0, 1));
  CuAssertIntEquals(tc, allKnown | LD,  mazeCell(0, 2));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_LEFT | KNOWN_DIRECTION_UP | LU,  mazeCell(0, 3));

  CuAssertIntEquals(tc, allKnown | DU,  mazeCell(1, 0));
  CuAssertIntEquals(tc, allKnown | DUR, mazeCell(1, 1));
  CuAssertIntEquals(tc, allKnown | DU,  mazeCell(1, 2));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_UP | DIRECTION_UP,  mazeCell(1, 3));

  CuAssertIntEquals(tc, allKnown | UR,  mazeCell(2, 0));
  CuAssertIntEquals(tc, allKnown | LR, mazeCell(2, 1));
  CuAssertIntEquals(tc, allKnown | DIRECTION_DOWN,  mazeCell(2, 2));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_UP | DIRECTION_UP,  mazeCell(2, 3));

  CuAssertIntEquals(tc, KNOWN_DIRECTION_UP | KNOWN_DIRECTION_LEFT | LU, mazeCell(3, 0));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_DOWN | KNOWN_DIRECTION_LEFT | LD, mazeCell(3, 1));
  CuAssertIntEquals(tc, allKnown | DU, mazeCell(3, 2));
  CuAssertIntEquals(tc, KNOWN_DIRECTION_UP | DIRECTION_UP, mazeCell(3, 3));
}

void createTestMaze() {
  addPercept(DIRECTION_DOWN | DIRECTION_RIGHT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_UP);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_UP);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_UP | DIRECTION_LEFT);

  mazeGet()->robot.x = 0;
  mazeGet()->robot.y++;
  addPercept(DIRECTION_RIGHT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_UP);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_UP);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_LEFT);

  mazeGet()->robot.x = 0;
  mazeGet()->robot.y++;
  addPercept(DIRECTION_RIGHT | DIRECTION_LEFT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_RIGHT | DIRECTION_DOWN);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_UP | DIRECTION_LEFT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_RIGHT | DIRECTION_LEFT);

  mazeGet()->robot.x = 0;
  mazeGet()->robot.y++;
  addPercept(DIRECTION_RIGHT | DIRECTION_LEFT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_RIGHT | DIRECTION_LEFT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_DOWN | DIRECTION_RIGHT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_LEFT);

  mazeGet()->robot.x = 0;
  mazeGet()->robot.y++;
  addPercept(DIRECTION_RIGHT | DIRECTION_UP);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_UP | DIRECTION_LEFT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_UP | DIRECTION_RIGHT);
  mazeGet()->robot.x++;
  addPercept(DIRECTION_UP | DIRECTION_LEFT);
}

void MazeTestFinishFind(CuTest* tc) {
  mazeInitialize();
  createTestMaze();

  unsigned int x, y;
  CuAssertTrue(tc, getFinishCoords(&x, &y));
  CuAssertIntEquals(tc, 2, x);
  CuAssertIntEquals(tc, 3, y);
}

CuSuite* MazeTestsGetSuite() {
  CuSuite* suite = CuSuiteNew();
  SUITE_ADD_TEST(suite, MazeTestInit);
  SUITE_ADD_TEST(suite, MazeTestInitWallCorners);
  SUITE_ADD_TEST(suite, MazeTestInitWallSides);
  SUITE_ADD_TEST(suite, MazeTestInitWallCenter);
  SUITE_ADD_TEST(suite, MazeTestPerceptRotation180);
  SUITE_ADD_TEST(suite, MazeTestPerceptRotation90);
  SUITE_ADD_TEST(suite, MazeTestPerceptRotation_90);
  SUITE_ADD_TEST(suite, MazeTestPerceptOrientationDown);
  SUITE_ADD_TEST(suite, MazeTestPerceptOrientationUp);
  SUITE_ADD_TEST(suite, MazeTestPerceptOrientationLeft);
  SUITE_ADD_TEST(suite, MazeTestPerceptOrientationRight);
  SUITE_ADD_TEST(suite, MazeTestAddingPercept);
  SUITE_ADD_TEST(suite, MazeTestAddingPercept2);
  SUITE_ADD_TEST(suite, MazeTestFinishFind);
  return suite;
}
